//==============================================================================
//
//  Copyright
//
//  José Villén-Altamirano <jose.villen@upm.es> (Universidad Politécnica de Madrid)
//
//  Contributor: Carlos E. Budde <c.e.budde@utwente.nl> (University of Twente)
//
//------------------------------------------------------------------------------
//
//  This file is part of the artifact created for the replication paper
//  "Replicating RESTART with Prolonged Retrials: An Experimental Report",
//  by Carlos E. Budde and Arnd Hartmanns (the "authors").
//
//  The source code was provided by José Villén-Altamirano (the "owner"),
//  and modified by the authors to fit the paper.
//
//  All lines starting with '//' were added by the authors.
//  The following comment codes are used:
//   - NOTE: comment for clarification
//   - PPP:  code added for printing (should not modify behaviour)
//   - XXX:  code modification (should not modify behaviour of RESTART)
//   - BUG:  code modification expected to modify behaviour of RESTART
//
//==============================================================================


#include <cstdio>
// XXX: changed stdlib.h to cstdlib
#include <cstdlib>
// XXX: added cassert for assert()
#include <cassert>
#include <cstring>
#include <cmath>
#include <ctime>

#include <string>
#include <iostream>
#include <fstream>
#include <sstream>
// XXX: added chrono for more precise time tracking
#include <chrono>

using namespace std;
using chrono::system_clock;
using time_lapse = chrono::duration<double>;

/*************************************************************************/
/*********************** PARAMETROS GENERALES ****************************/
/*************************************************************************/

#define TRUE  1
#define FALSE 0
#define MAX_L 4096

// https://stackoverflow.com/a/3782064
#define NO_OTHER_MACRO_STARTS_WITH_THIS_NAME_
#define IS_EMPTY(name) defined(NO_OTHER_MACRO_STARTS_WITH_THIS_NAME_ ## name)
#define EMPTY(name) IS_EMPTY(name)

// XXX: RNG seed
#if !defined(RNGSEED) || EMPTY(RNGSEED)
#  error "Must define RNGSEED at compile time, e.g. -DRNGSEED=42"
#endif

// XXX: choose model to run    (@compile: make CFLAGS=-DMODEL=3)
// NAME | MODEL:
//   2  | 2-tandem queue  (Markovian)
//   3  | 3-tandem queue  (Markovian)
//  33  | 3-tandem queue  (non-Markovian)
//   7  | 7-nodes network (Markovian)
//  77  | 7-nodes network (non-Markovian)
#if !defined(MODEL) || EMPTY(MODEL)
#  error "Must define a MODEL at compile time, e.g. -DMODEL=2"
#elif MODEL != 2 \
 && MODEL != 3   \
 && MODEL != 33  \
 && MODEL != 7   \
 && MODEL != 77
#  error "Invalid MODEL chosen, must be one of: 2, 3, 33, 7, 77"
#endif

// XXX: choose RESTART prolongation   (@compile: make CFLAGS=-DPj=2)
#if !defined(Pj) || EMPTY(Pj)
#  error "Must define Pj RESTART prolongation at compile time, e.g. -DPj=1"
#elif Pj != 0 \
 && Pj != 1 \
 && Pj != 2 \
 && Pj != 3 \
 && Pj != 4
#  error "Invalid Pj chosen (RESTART prolongation), must be one of 0, 1, 2, 3, 4"
#endif

/*************************************************************************/
/*********************** PARAMETROS DE SIMULACION ************************/
/*************************************************************************/

#define Num_Simulaciones 1 // NUMERO DE SIMULACIONES A COMPARAR
#define Num_Ejecuciones 1 // NUMERO DE EJECUCIONES A REALIZAR

// XXX Models implemented by Carlos E. Budde based on the paper
//     "RESTART vs. Splitting: A comparative study,"
//     by J. Villen-Altamirano, Performance Evaluation 121-122 (2018)
#if MODEL == 2
//	2-tandem queue (Markovian)
#	define Num_Servidores 2
	const int Tasa_Llegada [Num_Servidores] = {1,0};
	const int Tasa_Servicio[Num_Servidores] = {2,3};
#elif MODEL == 3
//	3-tandem queue (Markovian)
#	define Num_Servidores 3
	const int Tasa_Llegada [Num_Servidores] = {2,0,0};
	const int Tasa_Servicio[Num_Servidores] = {3,4,6};
#elif MODEL == 7
//	7-nodes network (Markovian)
#	define Num_Servidores 7
	const int Tasa_Llegada [Num_Servidores] = {1,1,1,1,1,1,1};
	const int Tasa_Servicio[Num_Servidores] = {9,9,9,9,14,14,17};
#elif MODEL == 33
//	3-tandem queue (non-Markovian)
#	define Num_Servidores 3
#	define NUM_PROB_LLEGADAS 2
#	define MU1 (7.0)
#	define MU2 (7.0/6.0)
#	define ERLANG_SHAPE 3
	const double Prob_Llegada[Num_Servidores][NUM_PROB_LLEGADAS] = {{0.5,1.0},{1.0,1.0},{1.0,1.0}};
	const double Tasa_Llegada[Num_Servidores][NUM_PROB_LLEGADAS] = {{MU1,MU2},{0.0,0.0},{0.0,0.0}};
	const int Tasa_Servicio[Num_Servidores] = {9, 12, 18};
#elif MODEL == 77
//	7-nodes network (non-Markovian)
#	define Num_Servidores 7
#	define NUM_PROB_LLEGADAS 2
#	define MU1 (3.5)
#	define MU2 (3.5/6.0)
#	define ERLANG_SHAPE 3
	const double Prob_Llegada[Num_Servidores][NUM_PROB_LLEGADAS] = {{0.5,1.0},{0.5,1.0},{0.5,1.0},{0.5,1.0},{0.5,1.0},{0.5,1.0},{0.5,1.0}};
	const double Tasa_Llegada[Num_Servidores][NUM_PROB_LLEGADAS] = {{MU1,MU2},{MU1,MU2},{MU1,MU2},{MU1,MU2},{MU1,MU2},{MU1,MU2},{MU1,MU2}};
	const int Tasa_Servicio[Num_Servidores] = {27,27,27,27,42,42,51};
#endif

/* MATRIZ DE TRANSICIONES ENTRE NODOS:                                  */
/* Probabilidades *acumuladas* de tránsito entre nodos-servidores.      */
/* Se pone una columna mas para guardar la probabilidad de que no vaya  */
/* a ningun servidor (probabilidad de salir de la red por ese servidor) */
const double prob_transito[Num_Servidores][Num_Servidores + 1] =
{
#if MODEL == 2
//	2-tandem queue
	{0.0, 0.0, 1.0},
	{1.0, 1.0, 1.0},
#elif MODEL == 3 || MODEL == 33
//	3-tandem queue
	{0.0, 0.0, 1.0, 1.0},
	{0.0, 0.0, 0.0, 1.0},
	{1.0, 1.0, 1.0, 1.0},
#elif MODEL == 7 || MODEL == 77
//	7-nodes network
	{0.2, 0.3, 0.4, 0.5, 0.6, 0.8, 1.0, 1.0},
	{0.2, 0.3, 0.4, 0.5, 0.6, 0.8, 1.0, 1.0},
	{0.2, 0.3, 0.4, 0.5, 0.6, 0.8, 1.0, 1.0},
	{0.2, 0.3, 0.4, 0.5, 0.6, 0.8, 1.0, 1.0},
	{0.2, 0.3, 0.4, 0.5, 0.6, 0.6, 0.7, 1.0},
	{0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.7, 1.0},
	{0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 1.0},
#endif
};

//------------------------------------------------------------------------------------------------------------------------------------
/* Características de la muestra */
#define MAXDOUBLE 100000000
#define Tam_Transitorio 40 /* Llegadas a partir de las cuales termina el transitorio*/
// NOTE: stop each RESTART simulation run when reloj >= Reloj_Final,
//       defined per model:
#if MODEL == 2
#	define Reloj_Final 30000
#elif MODEL == 3
#	define Reloj_Final 3000
#elif MODEL == 33
#	define Reloj_Final 3000
#elif MODEL == 7
#	define Reloj_Final 500
#elif MODEL == 77
#	define Reloj_Final 500
#endif

//------------------------------------------------------------------------------------------------------------------------------------
/* Condiciones fin de simulación */
#define Max_Muestras 25000 /* Límite por si no se logra un error relativo aceptable en un numero finito de muestras*/
#define Calcula_Error 5   /* Número de muestras a partir de las que empieza a calcular el error relativo */
#if MODEL == 2
#	define Max_Error 0.005     /* Error relativo MINIMO para el cual termina la simulación */
#elif MODEL == 3
#	define Max_Error 0.1
#elif MODEL == 33
#	define Max_Error 0.1
#elif MODEL == 7
#	define Max_Error 0.1
#elif MODEL == 77
#	define Max_Error 0.1
#endif
/* Valor de t-de-Student para el calculo del error */
const double Ta[56] = {2.36,2.36,2.36,2.36,2.36,2.36,2.36,2.36,2.31,2.26
					  ,2.23,2.20,2.18,2.16,2.14,2.13,2.12,2.11,2.10,2.09
					  ,2.09,2.08,2.07,2.07,2.06,2.06,2.06,2.05,2.05,2.04
					  ,2.04,2.04,2.04,2.04,2.03,2.03,2.03,2.03,2.02,2.02
					  ,2.02,2.02,2.02,2.02,2.02,2.02,2.01,2.01,2.01,2.01
					  ,2.01,2.01,2.01,2.01,2.01,2.01
					  };

//------------------------------------------------------------------------------------------------------------------------------------
/* PROBABILIDAD A CALCULAR                            */
/* Se calcula la siguiente probabilidad:              */
/*       P(Servidor_Simulado >= Clientes_En_Servidor) */
#if MODEL == 2
//	2-tandem queue
#	define Servidor_Simulado 2     /* Servidor del que se realiza la simulación */
#	define Clientes_En_Servidor 30 /* Num de clientes para calcular la probabilidad */
#elif MODEL == 3
//	3-tandem queue (Markovian)
#	define Servidor_Simulado 3
#	define Clientes_En_Servidor 30
#elif MODEL == 33
//	3-tandem queue (non-Markovian)
#	define Servidor_Simulado 3
#	define Clientes_En_Servidor 14
#elif MODEL == 7
//	7-nodes network (Markovian)
#	define Servidor_Simulado 7
#	define Clientes_En_Servidor 30
#elif MODEL == 77
//	7-nodes network (non-Markovian)  -->  same as Markovian:
#	define Servidor_Simulado 7
#	define Clientes_En_Servidor 30
#endif

/*************************************************************************/
/*********************** PARAMETROS DEL METODO RESTART *******************/
/*************************************************************************/

/* Se considera que el umbral de la primera zona es 0 y se incluye en el array */
// NOTE: choose always '0' as the threshold of the first zone, and '1' as its splitting
// XXX: make the following match the values of the reproduced papers
#if MODEL == 2
//	2-tandem queue
#	define Numero_Umbrales 28
	const int Umbrales    [Numero_Umbrales + 1] = {0,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29};
	const int Repeticiones[Numero_Umbrales + 1] = {1,3,2,2,3,3,3,2,3, 3, 3, 3, 3, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 5};
#elif MODEL == 3
//	3-tandem queue (Markovian)
#	define Numero_Umbrales 33
	const int Umbrales    [Numero_Umbrales + 1] = {0,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34};
	const int Repeticiones[Numero_Umbrales + 1] = {1,2,2,2,2,2,3,2,3, 3, 3, 2, 3, 3, 3, 2, 3, 3, 3, 2, 3, 3, 3, 3, 3, 2, 3, 3, 3, 3, 2, 3, 3, 3};
#elif MODEL == 33
//	3-tandem queue (non-Markovian)
#	define Numero_Umbrales 29
	const double Umbrales [Numero_Umbrales + 1] = {0.0, 1.5, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0, 5.5, 6.0, 6.5, 7.0, 7.5, 8.0, 8.5, 9.0, 9.5, 10.0, 10.5, 11.0, 11.5, 12.0, 12.5, 13.0, 13.5, 14.0, 14.5, 15.0, 15.5, 16.0};
	const int Repeticiones[Numero_Umbrales + 1] = {  1,   2,   3,   2,   3,   2,   3,   3,   3,   3,   3,   3,   3,   3,   3,   3,   3,    3,    3,    3,    3,    3,    3,    3,    3,    4,    3,    3,    4,    5};
#elif MODEL == 7
//	7-nodes network (Markovian)
#	define Numero_Umbrales 31
	const int Umbrales    [Numero_Umbrales + 1] = {0,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32};
	const int Repeticiones[Numero_Umbrales + 1] = {1,2,2,3,3,3,3,3,3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2};
#elif MODEL == 77
//	7-nodes network (non-Markovian)
#	define Numero_Umbrales 31
	const int Umbrales    [Numero_Umbrales + 1] = {0,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32};
	const int Repeticiones[Numero_Umbrales + 1] = {1,2,2,3,2,3,3,3,2, 3, 3, 3, 3, 2, 3, 3, 3, 2, 4, 3, 3, 2, 3, 3, 3, 3, 3, 3, 3, 3, 2, 4};
#endif

//------------------------------------------------------------------------------------------------------------------------------------
/* VALOR DE 'a' PARA EL CALCULO DE LA IMPORTANCIA */
#if MODEL == 2
//	2-tandem queue (Markovian)
	const double a[Num_Servidores] = {0.63, 1.0};
#elif MODEL == 3
//	3-tandem queue (Markovian)
	const double a[Num_Servidores] = {0.37, 0.63, 1.0};
#elif MODEL == 33
//	3-tandem queue (non-Markovian)
	const double a[Num_Servidores] = {0.13, 0.5, 1.0};
#elif MODEL == 7
//	7-nodes network (Markovian)
	const double a[Num_Servidores] = {0.23, 0.23, 0.23, 0.23, 0.45, 0.45, 1.0};
#elif MODEL == 77
//	7-nodes network (non-Markovian)  -->  same as Markovian:
	const double a[Num_Servidores] = {0.23, 0.23, 0.23, 0.23, 0.45, 0.45, 1.0};
#endif

//------------------------------------------------------------------------------------------------------------------------------------

double Media_Pi[Numero_Umbrales + 1][Num_Simulaciones];
double ri[Numero_Umbrales][Num_Simulaciones];
int Ri[Numero_Umbrales][Num_Simulaciones];
double extremo_inf[Num_Simulaciones];
double extremo_sup[Num_Simulaciones];
double Error_Relativo[Num_Simulaciones]; /* Calculo del error relativo */
int Van_Num_Muestras[Num_Simulaciones];   /* Número de muestras que se han realizado */
double Media[Num_Simulaciones]; /* Media de la muestras (parámetro a estimar) */
int NumVecesSinClientes[Num_Simulaciones]; /* Numero de veces que se queda sin Clientes */
int NumVecesImportanciaCero; /* para su acumulacion parcial */

int VanPosicion; /* 0:RESTART , 1:SPLITTING (deprecated) , 2:SPLITTING-CORTADO (deprecated) */

/*****************************************************************************/
/*                       DECLARACION DE TIPOS                                */
/*****************************************************************************/

//------------------------------------------------------------------------------------------------------------------------------------
/*** Definición de variables de cada servidor *******/
typedef enum {
	Serv_Ocupado,
	Serv_Libre
} estado_serv;

typedef struct servidor {
	/******* Variables de estado del servidor  ****************/
	estado_serv  Estado;      	/* Estado en el que se encuentra          */
	int          Num_Clientes;   	/* Peticiones en el servidor              */
	double       Tiempos[Clientes_En_Servidor + 1];   /* Contadores de tiempo */
	double       Cliente;      /* Peticion que esta siendo servida */
	/********   Variables de tiempos de sucesos  **************/
	double       Llegada;     /* Proxima llegada al servidor                */
	double       Salida;      /* Proxima salida del servidor                */
	double       Suceso_Ant;  /* Reloj del ultimo suceso del servidor       */
	/***********    Variables estadisticas       ***************/
	long         Clientes_Servidos;
	double       Tiempo_Cola;
} Tipo_Servidor;

//------------------------------------------------------------------------------------------------------------------------------------

/***********  Definicion de de contadores de tiempo de cada umbral *******/
typedef struct tiempos_umbrales {
	//+2 para que no de error raro???
	double Tiempos[Numero_Umbrales + 1];/* tiempo que pasa en cada umbral */
	double reloj;     					 /* tiempo global de la simulacion */
	double suceso_ant; 				 /* tiempo en el que se entro en el umbral actual */
	int umbral_calculo[Numero_Umbrales + 1];
} Tipo_Tiempo_Umbral;

/*****************************************************************************/
/*                   DECLARACION DE VARIABLES globales                       */
/*****************************************************************************/

double reloj;
// PPP: added the variable "TiempoTotalTransitorio"
double TiempoTotalTransitorio;  // Total time spent in transient phase
double Probabilidades[Max_Muestras]; /* Resultado simulado de cada muestra */
double Probabilidades_Rep[Max_Muestras][Numero_Umbrales + 1]; /* Resultado de la probabilidad Ti */

/*****************************************************************************/
/*                        FUNCIONES GENERICAS                                    */
/*****************************************************************************/

/*****************************************************************************/
double Calcula_Importancia(Tipo_Servidor Servidores[])
{
	double Calculado = 0;
  	for (int i = 0; i < Num_Servidores; i++)
	{
		Calculado = Calculado + (a[i] * Servidores[i].Num_Clientes);
	}

	if (Calculado == 0)
	{
		NumVecesImportanciaCero++;
	}

	return Calculado;
}

/*****************************************************************************/
static inline double Unif01() { return rand() / static_cast<double>(RAND_MAX); }

/*****************************************************************************/
//	XXX: updated for efficiency
static inline double
Exponencial(const double lambda)
{
	__volatile__ double u;
	do {
		u = Unif01();
	} while (u <= 0.0);
	return lambda <= 0.0 ? INFINITY
						 : -log(u)/lambda;
}

/*****************************************************************************/
// XXX: added for the arrivals of the non-Markovian models
static inline double
HiperExponencial(const unsigned N,
                 const double p[],
                 const double lambda[])
{
	static constexpr double EPSILON = 1.0e-6;
	assert(0u < N);
	assert(nullptr != p);
	assert(nullptr != lambda);
	assert(0.0 <= p[0]);
	assert((std::abs(1.0-p[N-1]) <= EPSILON * std::abs(p[N-1])));
	// ^^^ p[] is a vector of accumulated probability values:
	//     0.0 <= p[0] <= p[i] <= p[i+1] <= p[N-1] == 1.0
	const double u = Unif01();
	for (uint i=0ul ; i<N ; i++)
		if (u < p[i])
			return Exponencial(lambda[i]);
	assert(false);
	return INFINITY;
}

/*****************************************************************************/
// XXX: updated for efficiency
static inline double
Erlang(const int a, const double b)
{
	assert(0 < a);
	volatile double acum = 1.0;
	for (int i=0 ; i<a ; i++)
		acum *= Unif01();
	return -log(acum) / b;
}

/*****************************************************************************/
double Student (int n)
{
	if (n <= 55)
		return Ta[n];
	else if ((n>=56) && (n<=80))
		return 2.0;
	else if ((n>=81) && (n<=105))
		return 1.99;
	else if ((n>=106) && (n<=130))
		return 1.98;
	else if ((n>=131) && (n<=200))
		return 1.97;
	else
		return 1.96;
}

/*****************************************************************************/
void Inicializa_Servidores(Tipo_Servidor Servidores[])
{
	int i,j;
  	for (i = 0; i < Num_Servidores; i++)
	{
		Servidores[i].Estado = Serv_Libre;
		Servidores[i].Num_Clientes = 0;
		for (j = 0; j < Clientes_En_Servidor + 1 ; j++)
			Servidores[i].Tiempos[j] = 0;
		Servidores[i].Cliente = -1.0;
// XXX: pragma-parameterised per-model distributions
#if MODEL == 2 || MODEL == 3 || MODEL == 7
		Servidores[i].Llegada = Exponencial(Tasa_Llegada[i]); // CALCULA LA PRIMERA LLEGADA
#elif MODEL == 33 || MODEL == 77
		Servidores[i].Llegada = HiperExponencial(NUM_PROB_LLEGADAS,Prob_Llegada[i],Tasa_Llegada[i]);
		//Servidores[i].Llegada = Exponencial(Tasa_Llegada[i]); // CALCULA LA PRIMERA LLEGADA
#endif
		Servidores[i].Salida = INFINITY;  /* número alto para la primera comparación */
		Servidores[i].Suceso_Ant = 0.0;
		Servidores[i].Clientes_Servidos = 0;
		Servidores[i].Tiempo_Cola = 0.0;
	}
}

/*****************************************************************************/
void Inicializa_Despues_de_Transitorio(Tipo_Servidor Servidores[],Tipo_Tiempo_Umbral* Tiempos_umbral)
{
	int i,j;
	for (i=0; i < Num_Servidores ;i++)
	{
		for (j=0; j < Clientes_En_Servidor + 1 ;j++)
			Servidores[i].Tiempos[j] = 0.0;
	}

	for (i=0; i <= Numero_Umbrales ;i++)
	{
		Tiempos_umbral->Tiempos[i] = 0.0;
		Tiempos_umbral->umbral_calculo[i] = 0;
	}
	Tiempos_umbral->reloj = 0.0;
	Tiempos_umbral->suceso_ant = 0.0;

//	PPP: added the following logging of the total time spent in transient phase
	TiempoTotalTransitorio = reloj;

//	BUG: the following was added so RESTART runs until 'RelojFinal';
//	     else it runs until 'RelojFinal-TiempoTotalTransitorio' time units,
//	     which is wrong given the calculations in 'Calcula_Probabilidad()'
	reloj = 0.0;
}

/*****************************************************************************/
/* MIRA en que servidor se realiza la proxima llegada y devuelve el tiempo en el que ocurre*/
/* Y EL SERVIDOR EN EL QUE OCURRE (serv) !!!! */
double Proxima_Llegada(int *serv,Tipo_Servidor Servidores[])
{
  double menor=MAXDOUBLE;
  int i;

  for (i = 0; i < Num_Servidores; i++)
	if (Servidores[i].Llegada < menor)
	{
	  *serv=i+1;
	  menor=Servidores[i].Llegada;
	}

  /** traza SI HAY ERROR Y ES MENOR QUE EL RELOJ ACTUAL !!!! */
  if (menor<reloj)
  {
  	printf("menor\n") ;
//	XXX: added the following forceful quit
	exit(EXIT_FAILURE);
  }
 return menor;
}

/*****************************************************************************/
/*MIRA en que servidor se realiza la proxima salida y devuelve el tiempo en el que ocurre*/
/* Y EL SERVIDOR EN EL QUE OCURRE (serv) !!!! */
double Proxima_Salida(int *serv,Tipo_Servidor Servidores[])
{
  double menor=MAXDOUBLE;
  int i;

// BUG: added braces to enclose the for-loop body
//  for (i = 0; i < Num_Servidores; i++)
  for (i = 0; i < Num_Servidores; i++) {
	if ((Servidores[i].Salida < menor) && (Servidores[i].Num_Clientes > 0))
	 {
		*serv = i+1;
		menor = Servidores[i].Salida;
	 }

// BUG: changed 'serv' to '*serv'
//	if (serv < 0)
	if (*serv < 0)
	{
// PPP: removed print, that is always true at initialisation
//	  printf("serv menor que 0\n") ;
	}
	 /****traza*/
	if (menor<reloj)
	{
	  printf("menor\n") ;
	}
  }
	return menor;
}

/*****************************************************************************/
/* Dado un servidor de salida gestionado,                   */
/* en base a la matriz de probabilidades prob_transito,     */
/* seleccionar un destino hacia donde va.                   */
/* Salida: Numero de servidor o "0" si va fuera del sistema */
int SeleccionaDestino(int serv)
{
  int i = 0,destino = -1;
  const double prob = Unif01();

  while ((destino == - 1)&&(i <= Num_Servidores))
	{
	  if (prob_transito[serv-1][i] >= prob)
		{
			destino = i;
		}

	  i++;
	}

  return destino;  /* destino 0 es que va fuera del sistema */
}

/*****************************************************************************/
double Calcula_Indice_Umbral(int factor)
{
  int i;
  double solucion = 1;

  for (i = 0; i <= factor; i++)
  {
  	if (VanPosicion == 0)
	 	solucion = solucion * Repeticiones[i];
//	XXX: added the following "else force quit"
	else
		exit(EXIT_FAILURE);
  }
  return solucion;
}

/*****************************************************************************/
void Gestiona_Llegada(int serv
					 ,int De_Fuera
					 ,int umbral_act
					 ,Tipo_Servidor Servidores[]
					 ,Tipo_Tiempo_Umbral* Tiempos_umbral)
{
	int Indice_Tiempos; /* En caso de que haya mas de Clientes_En_Servidor en el nodo         */
	                 /* el tiempo Se le suma al mayor, ya que la probabilidad va ser ">="  */

	double indice_umbral; /* Limita en caso de que el num de clientes sea mayor que el que se quiere calcular en la probabilidad */

//	NOTE: next is: Indice_Tiempos = min(Clientes_en_Servidor, Servidores[serv-1].Num_Clientes)
	if (Servidores[serv-1].Num_Clientes > Clientes_En_Servidor)
		Indice_Tiempos = Clientes_En_Servidor;
	else
		Indice_Tiempos = (int)Servidores[serv-1].Num_Clientes;

	/*Actualiza reloj si la llegada es de fuera(si no ya se ha actualizado en salidas)*/
	if (De_Fuera)
	{
		/*actualiza el tiempo global de simulacion*/
		/*es el tiempo del suceso actual menos el del ultimo que esta en reloj*/
		/*nuevo*/
		Tiempos_umbral->reloj += Servidores[serv-1].Llegada - reloj;

		/*actualiza el reloj*/
		reloj = Servidores[serv-1].Llegada;
	}

	/******traza**/
	/*		printf("Llegada a %i en %f \n",serv,reloj);*/

	/* Calcular Ti nuevo */
	indice_umbral = Calcula_Indice_Umbral(umbral_act);
	if (indice_umbral == 0) /*****traza*/
	{
		indice_umbral = Calcula_Indice_Umbral(umbral_act);
	}

	Servidores[serv-1].Tiempos[Indice_Tiempos] +=
			(reloj - Servidores[serv-1].Suceso_Ant)/indice_umbral;

	/*Si el servidor esta libre pasa a cliente,calcula salida,Actualiza estado*/
	if (Servidores[serv-1].Estado == Serv_Libre)
 	{
		Servidores[serv-1].Estado = Serv_Ocupado;
		Servidores[serv-1].Cliente = reloj;
//		XXX: pragma-parameterised per-model distributions
#		if MODEL == 2 || MODEL == 3 || MODEL == 7
			Servidores[serv-1].Salida = reloj + Exponencial(Tasa_Servicio[serv-1]);
#		elif MODEL == 33 || MODEL == 77
			Servidores[serv-1].Salida = reloj + Erlang(ERLANG_SHAPE,Tasa_Servicio[serv-1]);
#		endif
	}

 	/* Incrementa Num_Clientes*/
	Servidores[serv-1].Num_Clientes++;

	/* Actualizar suceso anterior */
	Servidores[serv-1].Suceso_Ant = reloj;

	/* Calcular proxima Llegada si la llegada fue de fuera*/
	if (De_Fuera)
	{
//		XXX: pragma-parameterised per-model distributions
#		if MODEL == 2 || MODEL == 3 || MODEL == 7
			Servidores[serv-1].Llegada = reloj + Exponencial(Tasa_Llegada[serv-1]);
#		elif MODEL == 33 || MODEL == 77
			Servidores[serv-1].Llegada = reloj + HiperExponencial(NUM_PROB_LLEGADAS,Prob_Llegada[serv-1],Tasa_Llegada[serv-1]);
#		endif
	}
}

/*****************************************************************************/
/* Gestiona la salida de un mensaje de un serv y calcula hacia donde va el siguiente mensaje */
void Gestiona_Salida(int serv
					,int umbral_act
					,Tipo_Servidor Servidores[]
					,Tipo_Tiempo_Umbral* Tiempos_umbral)
{
	int destino;
	double saliente;
	int Indice_Tiempos; /* En caso de que haya mas de Clientes_En_Servidor en el nodo         */
						/* el tiempo se le suma al mayor, ya que la probabilidad va ser ">="  */

	double indice_umbral;/* Limita en caso de que el num de clietes sea mayor que el que se quiere calcular en la probabilidad */
						 /* numero que divide el tiempo en funcion del umbral en el que este*/

//	NOTE: next is: Indice_Tiempos = min(Clientes_en_Servidor, Servidores[serv-1].Num_Clientes)
	if (Servidores[serv-1].Num_Clientes > Clientes_En_Servidor)
		Indice_Tiempos = Clientes_En_Servidor;
	else
		Indice_Tiempos = (int)Servidores[serv-1].Num_Clientes;

 	saliente = Servidores[serv-1].Cliente;
//	NOTE: 'serv' is used as 'serv-1' inside SeleccionaDestino()
 	destino = SeleccionaDestino(serv);

	/*Actualiza el reloj global y el parcial de simulacion*/
	/*nuevo*/
	(*Tiempos_umbral).reloj = (*Tiempos_umbral).reloj + (Servidores[serv-1].Salida - reloj);
	reloj = Servidores[serv-1].Salida;

	/*******traza******/
	/*	 printf("Salida de %i a %i en %f  \n",serv,destino,reloj);*/

 	/*Actualiza el tiempo que el cliente está en cola (si sacas a alguien de la cola)*/
	/*Actualiza contadores de tiempo Ti nuevo */
	indice_umbral=Calcula_Indice_Umbral(umbral_act);
	if (indice_umbral == 0)/*****traza*/
	{
		indice_umbral=Calcula_Indice_Umbral(umbral_act);
	}

	Servidores[serv-1].Tiempos[Indice_Tiempos] = Servidores[serv-1].Tiempos[Indice_Tiempos]
	 								+ ((reloj - Servidores[serv-1].Suceso_Ant)/indice_umbral);

	/*Actualiza numclientes */
	Servidores[serv-1].Num_Clientes--;

	/*Actualiza Proxima salida con tiempo de servicio del servidor(si hay cliente) */
	if (Servidores[serv-1].Num_Clientes > 0)
	{
//		XXX: pragma-parameterised per-model distributions
#		if MODEL == 2 || MODEL == 3 || MODEL == 7
			Servidores[serv-1].Salida = reloj + Exponencial(Tasa_Servicio[serv-1]);
#		elif MODEL == 33 || MODEL == 77
			Servidores[serv-1].Salida = reloj + Erlang(ERLANG_SHAPE,Tasa_Servicio[serv-1]);
#		endif
	}
	else
	{
		Servidores[serv-1].Estado=Serv_Libre;
	}

	/*Actualiza Suceso Anterior*/
	Servidores[serv-1].Suceso_Ant = reloj;

	/*Actualiza clientes servidos*/
	Servidores[serv-1].Clientes_Servidos++;

	/*Si procede, gestionar entrada al servidor destino*/
	if (destino!=0)
	{
		int Viene_De_Fuera = FALSE;
		Gestiona_Llegada(destino
						,Viene_De_Fuera
						,umbral_act
						,Servidores
		                ,Tiempos_umbral);
	}
}

/*****************************************************************************/
double Tiempo_Restart(Tipo_Tiempo_Umbral* Tiempos_umbral)
{
 double solucion=0,indice_umbral;
 int i;

	for (i=0;i <= Numero_Umbrales;i++)
	{
		indice_umbral = Calcula_Indice_Umbral(i);
		solucion += (Tiempos_umbral->Tiempos[i] / indice_umbral);
	 }

	 return solucion;
}

/*****************************************************************************/
double Calcula_Probabilidad(Tipo_Servidor Servidores[]
							,Tipo_Tiempo_Umbral* Tiempos_umbral)
{
	double tiempo;

	tiempo = Tiempo_Restart(Tiempos_umbral);
	tiempo = Reloj_Final;

	return Servidores[Servidor_Simulado-1].Tiempos[Clientes_En_Servidor] / tiempo;
}

/*****************************************************************************/
double Calcula_Probabilidad_Ti(Tipo_Tiempo_Umbral* Tiempos_umbral
							,int umbral)
{
	double tiempo;
	double indice_umbral;
	double suma_tiempos = 0;
	int i;

 	tiempo = Tiempo_Restart(Tiempos_umbral);
	//i=umbral porque los umbrales van de 0 a n (el 0 está incluido)
	for (i = umbral; i <= Numero_Umbrales; i++)
	{
		indice_umbral = Calcula_Indice_Umbral(i);
		suma_tiempos += (*Tiempos_umbral).Tiempos[i]/indice_umbral;
	}
 	return suma_tiempos / tiempo;
}

/*****************************************************************************/
double Calcula_Media(int n)
{
  int i;
  double suma=0;

  for (i=0;i<n;i++)
	 suma += Probabilidades[i];

  return suma/(double)n;
}

/*****************************************************************************/
double Calcula_Media_Umbral(int n,int umbral)
{
  int i;
  double suma=0;

  for (i=0;i<n;i++)
	 suma += Probabilidades_Rep[i][umbral-1];

  return suma/(double)n;
}

/*****************************************************************************/
double Calcula_s2(int n,double med)
{
	int i;
	double suma=0;

  	for (i=0;i<n;i++)
		suma += (Probabilidades[i]-med)*(Probabilidades[i]-med);

  	return suma/(n-1);
}

/*****************************************************************************/
double Calcula_Error_Relativo(double desv,double med,int n)
{
	return (Student(n) * (desv / sqrt(n)))/med;
}

/*****************************************************************************/
double Calcula_Intervalo_Confianza(double desv,int n)
{
	return (Student(n) * (desv / sqrt(n)) );
}

/*****************************************************************************/
void Imprime_Pantalla(int n, double Error, double med)
{
//	PPP: Changed the following log print:
	//printf("En %i muestras, error de %.4f con media %e  \n", n, Error, med);
	printf("[%4d]  Mean: %.2e  Rel.Err: %.4f  Transtime: %.1f\n",
		   n,
		   med,
		   Error,
		   TiempoTotalTransitorio);
	fflush(stdout);
}

/*****************************************************************************/
void Transitorio(Tipo_Servidor Servidores[],Tipo_Tiempo_Umbral* Tiempos_umbral)
{
	int servidor; /* Servidor en el que ocurre el siguiente evento */
	int serv_salida = -1;
	int serv_entrada = -1;

	reloj = 0.0;
	long Clientes_Servidos = 0; /* Total de clientes servidos por el sistema */

	Inicializa_Servidores(Servidores);
	/*Muestra_Servidores();*/
	int i;
	for (i=0; i <= Numero_Umbrales ;i++)
		{
			Tiempos_umbral->Tiempos[i] = 0.0;
			Tiempos_umbral->umbral_calculo[i] = 0;
		}
	Tiempos_umbral->reloj = 0.0;
	Tiempos_umbral->suceso_ant = 0.0;

	int Viene_De_Fuera = FALSE;

  	do
 	{
 	    // SE CALCULA LA PROXIMA LLEGADA Y SALIDA Y SE MIRA A VER CUAL SE VA A DAR ANTES
		if (Proxima_Salida(&serv_salida,Servidores) <= Proxima_Llegada(&serv_entrada,Servidores))
		{
		  // Si la Proxima_Salida es menor o igual que la Proxima_Llegada HACE SALIDA
		  servidor=serv_salida;
		  Gestiona_Salida(servidor
						  ,0
						  ,Servidores
		                  ,Tiempos_umbral);
		}
		else
		{
		  // Si la Proxima_Salida es MAYOR que la Proxima_Llegada HACE LLEGADA
		  servidor=serv_entrada;
		  Viene_De_Fuera = TRUE;
		  Gestiona_Llegada(servidor
		  					,Viene_De_Fuera
							,0
							,Servidores
		                    ,Tiempos_umbral);
		  Clientes_Servidos++;
		}

	} while (Clientes_Servidos < Tam_Transitorio);

 	Inicializa_Despues_de_Transitorio(Servidores,Tiempos_umbral);
}

/*****************************************************************************/
// XXX: created to differentiate between Markovian and non-Markovian
//      restoration of the previous state
// Restore servers to previous (run) values,
// after a RESTART trial has been killed by going down too many levels
void Restaurar_Servidores(Tipo_Servidor Servidores[],
						  Tipo_Servidor Servidores_Guardados[])
{
	for (uint i = 0u ; i < Num_Servidores ; i++) {
		Servidores[i].Estado       = Servidores_Guardados[i].Estado;
		Servidores[i].Num_Clientes = Servidores_Guardados[i].Num_Clientes;
		Servidores[i].Cliente      = Servidores_Guardados[i].Cliente;
		Servidores[i].Suceso_Ant   = Servidores_Guardados[i].Suceso_Ant;
#		if MODEL == 2 || MODEL == 3 || MODEL == 7
			// For Markovian systems re-schedule arrivals and departures
			Servidores[i].Llegada = reloj + Exponencial(Tasa_Llegada[i]);
			if (Servidores[i].Num_Clientes > 0)
				Servidores[i].Salida = reloj + Exponencial(Tasa_Servicio[i]);
#		elif MODEL == 33 || MODEL == 77
			// For non-Markovian systems just recover them from the stored state
			Servidores[i].Llegada  = Servidores_Guardados[i].Llegada;
			Servidores[i].Salida   = Servidores_Guardados[i].Salida;
#		endif
	}
}

/*****************************************************************************/
void AcumulaTiemposUmbral(int Umbral, Tipo_Tiempo_Umbral* Tiempos_umbral)
{
	Tiempos_umbral->Tiempos[Umbral] += Tiempos_umbral->reloj - Tiempos_umbral->suceso_ant;
	Tiempos_umbral->suceso_ant = Tiempos_umbral->reloj;
}

/*****************************************************************************/
void Restart(int umbral_act
			,Tipo_Servidor Servidores[]
			,Tipo_Tiempo_Umbral* Tiempos_umbral)
{
	int servidor; /* Servidor en el que ocurre el siguiente evento */
	int serv_salida = -1;
	int serv_entrada = -1;
	int i;
	// PARA GUARDAR el estado con el que entra al umbral Y LUEGO PODER RECUPERARLO
	Tipo_Servidor Servidor_Guardado[Num_Servidores];
	// PARA GUARDAR EL RELOJ CON EL QUE SE ENTRA AL UMBRAL Y LUEGO PODER RECUPERARLO
	double reloj_Guardado = 0;

	double importancia; /* guarda el calculo de la funcion de importancia */
	int repeticion = 1;   /* numero de repeticiones que se realizan en el umbral actual*/

 	/* guardo el estado actual */
	for (i = 0; i < Num_Servidores; i++)
	{
		Servidor_Guardado[i].Estado       = Servidores[i].Estado;
		Servidor_Guardado[i].Num_Clientes = Servidores[i].Num_Clientes;
		Servidor_Guardado[i].Cliente      = Servidores[i].Cliente;
		Servidor_Guardado[i].Suceso_Ant   = Servidores[i].Suceso_Ant;
		Servidor_Guardado[i].Llegada      = Servidores[i].Llegada;
		Servidor_Guardado[i].Salida       = Servidores[i].Salida;
	}
	reloj_Guardado = reloj;
	int Viene_De_Fuera = FALSE;

	// BUCLE PRINCIPAL HASTA QUE:
	// se alcance el numero de repeticiones por umbral
	do
 	{
 	    // SE CALCULA LA PROXIMA LLEGADA Y SALIDA Y SE MIRA A VER CUAL SE VA A DAR ANTES
		if (Proxima_Salida(&serv_salida,Servidores) <= Proxima_Llegada(&serv_entrada,Servidores))
		{
		  // SI LA Proxima_Salida es menor o igual que la Proxima_Llegada HACE SALIDA
		  servidor=serv_salida;
		  Gestiona_Salida(servidor
		  				  ,umbral_act
						  ,Servidores
		                  ,Tiempos_umbral);
		}
		else
		{
		  // SI LA Proxima_Salida es MAYOR que la Proxima_Llegada HACE LLEGADA
		  servidor=serv_entrada;
		  Viene_De_Fuera = TRUE;
		  Gestiona_Llegada(servidor
		  					,Viene_De_Fuera
						    ,umbral_act
							,Servidores
		                    ,Tiempos_umbral);
		}

		/* calculo la funcion de importancia */
		// EN FUNCION DEL NUMERO DE CLIENTES EN CADA SERVIDOR
		importancia = Calcula_Importancia(Servidores);

		/* en funcion de la importancia cambio o no de umbral*/
		if ((umbral_act < Numero_Umbrales)
			 && (importancia > Umbrales[umbral_act + 1])
		   )
		{
			// SI NO HE LLEGADO AL NUMERO TOTAL DE UMBRALES
			// Y
			// LA IMPORTANCIA CALCULADA ES MAYOR QUE LA DEL UMBRAL SIGUIENTE
			/* Sube de umbral*/
			AcumulaTiemposUmbral(umbral_act,Tiempos_umbral);
			/*Llamada recursiva para el siguiente nivel*/
			Restart(umbral_act + 1, Servidores, Tiempos_umbral);
		}
		else
		{
// XXX: choose j-prolongation of RESTART-Pj
#if   Pj == 0  // RESTART-P0,  aka "RESTART"
		if ( (  (umbral_act  > 0) && (importancia < Umbrales[umbral_act]) )
#elif Pj == 1  // RESTART-P1
		if (  ( (umbral_act == 1) && (importancia < Umbrales[umbral_act  ]) )
		   || ( (umbral_act  > 1) && (importancia < Umbrales[umbral_act-1]) )
#elif Pj == 2  // RESTART-P2
		if ( (  (umbral_act == 1) && (importancia < Umbrales[umbral_act  ]) )
		   || ( (umbral_act == 2) && (importancia < Umbrales[umbral_act-1]) )
		   || ( (umbral_act  > 2) && (importancia < Umbrales[umbral_act-2]) )
#elif Pj == 3  // RESTART-P3
		if ( (  (umbral_act == 1) && (importancia < Umbrales[umbral_act  ]) )
		   || ( (umbral_act == 2) && (importancia < Umbrales[umbral_act-1]) )
		   || ( (umbral_act == 3) && (importancia < Umbrales[umbral_act-2]) )
		   || ( (umbral_act  > 3) && (importancia < Umbrales[umbral_act-3]) )
#elif Pj == 4  // RESTART-P4
		if ( (  (umbral_act == 1) && (importancia < Umbrales[umbral_act  ]) )
		   || ( (umbral_act == 2) && (importancia < Umbrales[umbral_act-1]) )
		   || ( (umbral_act == 3) && (importancia < Umbrales[umbral_act-2]) )
		   || ( (umbral_act == 4) && (importancia < Umbrales[umbral_act-3]) )
		   || ( (umbral_act  > 4) && (importancia < Umbrales[umbral_act-4]) )
#endif
				)
			{
				// SI LA IMPORTANCIA CALCULADA ES MENOR QUE LA DEL UMBRAL ACTUAL
				// VER SI SE HA LLEGADO AL NUMERO DE REPETICIONES DEL UMBRAL
				/* Baja DEL umbral ACTUAL*/
				AcumulaTiemposUmbral(umbral_act,Tiempos_umbral);

				// SI NO HE LLEGADO AL NUMERO DE REPETICIONES DEL UMBRAL
				// recuperamos reloj y estado servidores AL ENTRAR AL UMBRAL
				// Y HACEMOS RE-SCHEDULING ANTES DE REPETIR (solo para Jackson)
				if (repeticion < Repeticiones[umbral_act])
				{
					reloj = reloj_Guardado;
					Restaurar_Servidores(Servidores, Servidor_Guardado);
				}
				repeticion++;
			}
		}

//	BUG: changed "Repeticiones1" for "Repeticiones" in the following:
	//} while (reloj < Reloj_Final && (repeticion <= Repeticiones1[umbral_act]));
	} while (reloj < Reloj_Final && (repeticion <= Repeticiones[umbral_act]));

	AcumulaTiemposUmbral(umbral_act,Tiempos_umbral);
}

///*****************************************************************************/
void Simulacion_Restart(int Posicion = 0)
{
	int i;
	double prob,intervalo_confianza;
	double s = 0;
	double RM = 1;

	// ARRAY CON LOS SERVIDORES A SIMULAR
	Tipo_Servidor Servidores[Num_Servidores]; // 7 = 0..6

	Tipo_Tiempo_Umbral Tiempos_umbral;

	/*Inicializo variable de probabilidades calculadas en cada muestra*/
	for (i = 0; i < Max_Muestras; i++)
		Probabilidades[i] = 0;

	// INICIALIZACION NUMERO DE MUESTRAS TRATADAS
	Van_Num_Muestras[Posicion] = 0;
	NumVecesSinClientes[Posicion] = 0;
	NumVecesImportanciaCero = 0;
	Error_Relativo[Posicion] = 1; /*  Para que cumpla la condicion las primeras veces*/

	do
	{
		Transitorio(Servidores,&Tiempos_umbral);

		// COMIENZA SIMULACION RESTART DESDE EL UMBRAL 0
		Restart(0,Servidores,&Tiempos_umbral);

		prob = Calcula_Probabilidad(Servidores,&Tiempos_umbral);
		Probabilidades[Van_Num_Muestras[Posicion]] = prob;

		// Probabilidades Ti
		for(i = 1; i <= Numero_Umbrales; i++)
		{
		   Probabilidades_Rep[Van_Num_Muestras[Posicion]][i-1] = Calcula_Probabilidad_Ti(&Tiempos_umbral,i);
		}

		// probabilidad TM+1
		Probabilidades_Rep[Van_Num_Muestras[Posicion]][Numero_Umbrales] = prob;

		for (i = 0; i < (Numero_Umbrales + 1); i++)
		{
		    Media_Pi[i][Posicion] = Calcula_Media_Umbral(Van_Num_Muestras[Posicion],i + 1);
	 	}

		Van_Num_Muestras[Posicion]++;
		if (Van_Num_Muestras[Posicion] > Calcula_Error)
		{
	  		Media[Posicion] = Calcula_Media(Van_Num_Muestras[Posicion]);
	  		s = sqrt(Calcula_s2(Van_Num_Muestras[Posicion], Media[Posicion]));
	  		Error_Relativo[Posicion] = Calcula_Error_Relativo(s, Media[Posicion], Van_Num_Muestras[Posicion]);
		}

		Imprime_Pantalla(Van_Num_Muestras[Posicion], Error_Relativo[Posicion], Media[Posicion]);

	} while ((Van_Num_Muestras[Posicion] < Max_Muestras) && (Error_Relativo[Posicion] > Max_Error));

	// Calculo ri
	for (i = 0; i < Numero_Umbrales; i++)
 	{
		ri[i][Posicion] = 1/sqrt((Media_Pi[i+1][Posicion]/Media_Pi[0][Posicion]) *  Media_Pi[i][Posicion] * 2);
	}
	ri[Numero_Umbrales-1][Posicion] = 1/sqrt((Media_Pi[Numero_Umbrales][Posicion]/Media_Pi[0][Posicion]) *  Media_Pi[Numero_Umbrales][Posicion] * 2);

	// Calculo Ri
	Ri[0][Posicion] = (int)round(ri[0][Posicion]);
  	for (i = 1; i < Numero_Umbrales; i++)
 	{
		RM = RM*(double)Ri[i-1][Posicion];
	 	Ri[i][Posicion] = (int)round((ri[i][Posicion]/RM));
	}

	NumVecesSinClientes[Posicion] = NumVecesImportanciaCero;

	printf("+------------------------------------------------------\n");
	printf("| Numero de simulaciones: %d\n", Van_Num_Muestras[Posicion]);
	printf("| Estimador puntual:      %.2e\n", Media[Posicion]);
	intervalo_confianza = Calcula_Intervalo_Confianza(s,Van_Num_Muestras[Posicion]);
 	extremo_inf[Posicion] = Media[Posicion] - intervalo_confianza;
 	extremo_sup[Posicion] = Media[Posicion] + intervalo_confianza;
	printf("| Intervalo de confianza: ( %.2e , %.2e )\n",
		   extremo_inf[Posicion],
		   extremo_sup[Posicion]);
}


/*****************************************************************************/
/*                           PROGRAMA PRINCIPAL                              */
/*****************************************************************************/

int main(void)
{
	int i,j, VanEjecuciones;

//	NOTE: Num_Ejecuciones == Num_Simulaciones == VanEjecuciones == 1, always

	for (VanEjecuciones = 1; VanEjecuciones <= Num_Ejecuciones; VanEjecuciones++)
	{

//	NOTE: init global vars
	for (j = 0; j < Num_Simulaciones; j++)
	{
		for (i = 0; i <= Numero_Umbrales; i++)
		{
		Media_Pi[i][j] = 0;
		}

		for (i = 0; i < Numero_Umbrales - 1; i++)
		{
		ri[i][j] = 0;
		Ri[i][j] = 0;
		}

		extremo_inf[j] = 0;
		extremo_sup[j] = 0;
		Error_Relativo[j] = 0; /* Calculo del error relativo */
		Media[j] = 0; /* Media de la muestras (parámetro a estimar) */
	}

//	PPP: changed the following log print
	const auto currentTime = system_clock::to_time_t(system_clock::now());
	printf("\n*** Programa comenzado @ %s\n", ctime(&currentTime));
	printf("+------------------------------------------------------\n");
#if   MODEL == 2
	printf("| Modelo: .............. 2-tandem queue (Markovian)\n");
#elif MODEL == 3
	printf("| Modelo: .............. 3-tandem queue (Markovian)\n");
#elif MODEL == 33
	printf("| Modelo: .............. 3-tandem queue (non-Markovian)\n");
#elif MODEL == 7
	printf("| Modelo: .............. 7-nodes network (Markovian)\n");
#elif MODEL == 77
	printf("| Modelo: .............. 7-nodes network (non-Markovian)\n");
#endif
	printf("| Num_Servidores:        %i\n", Num_Servidores);
	printf("| Tam_Transitorio: ..... %i\n", Tam_Transitorio);
	printf("| Servidor_Simulado:     %i\n", Servidor_Simulado);
	printf("| Clientes_En_Servidor:  %i\n", Clientes_En_Servidor);
	printf("| Numero_Umbrales:       %i\n", Numero_Umbrales);
	printf("| Max_Muestras: ........ %i\n", Max_Muestras);
	printf("| Max_Error:             %.3f\n", Max_Error);
	printf("| Tmax simulacion:       %i\n", Reloj_Final);
	printf("| Semilla para el RNG:   %i\n", RNGSEED);
	printf("| Calcula Rel.Err luego de la muestra [%i]\n",Calcula_Error);
	printf("+--------------------  RESTART-P%i  --------------------\n", Pj);

	// INICIALIZA CON EL NUMERO DE MILISEGUNDOS QUE LLEVA ENCENDIDO EL EQUIPO EL GENERADOR DE NUMEROS ALEATORIOS
	// QUE SERA DIFERENTE PARA CADA EJECUCION DEL PROGRAMA
//	XXX: changed to user-defined RNG seed, to make experiments reproducible
//	srand(time(NULL));
	srand(RNGSEED);

	const auto startTime = system_clock::now();

//	NOTE: entry point for all computations              <-----------------
	Simulacion_Restart();

	const time_lapse totalTime = system_clock::now() - startTime;
	printf("| Tiempo ejecucion (s):   %.2f\n", totalTime.count());
	printf("+------------------------------------------------------\n");

#ifdef DEBUG
	printf("Umbral\t");
	printf("Rep.\t");
	printf("Prob.\n");
	for (i = 0; i < Numero_Umbrales; i++)
 	{
		printf("%i\t",i + 1);  // Umbral
 		for (j = 0; j < Num_Simulaciones; j++)
 		{
			printf("%i\t", Ri[i][j]);  // Repeticiones
			printf("%.2e\n", Media_Pi[i][j] * 2);  // Probabilidad
		}
	}

 	printf("Intervalo de confianza;");
	printf("RESTART;");
	printf(";");

	printf("\n");
	printf(";");
	for (j = 0; j < Num_Simulaciones; j++)
	{
 		printf("( %e , %e );", extremo_inf[j],extremo_sup[j]);
		printf(";");
 	}
	printf("\n");
#endif

 	} // DE VanEjecuciones

	printf("\n*** Programa finalizado, tenga un buen dia ***\n");
}
