﻿using System;
using System.Runtime.CompilerServices;
using Modest.Exploration;
using CI = System.Globalization.CultureInfo;

namespace Modest.Simulation
{
	sealed class RestartRunManagerOldVersion<T>: RunManager<T>
		where T : struct, IState<T>
	{
		private readonly ImportanceFunction<T> importanceFunction;
		private readonly PropertyRunEvaluator<T> propertyEvaluator;
		private readonly int prolongation = 0;
		private readonly LongRunAverageValueProperty.LongRunAverageValueRunEvaluator<T> lraPropertyEvaluator;

		private double result;
		private int count;
		private RunState<T>[] agenda = new RunState<T>[16]; // per run
		private int[] currentLevel = new int[16]; // per run
		private int[] creationLevel = new int[16]; // per run
		private int[] splitLevel = new int[16]; // per run
		private int initialStateLevel;

		public RestartRunManagerOldVersion(SimulationRun<T> run, ImportanceFunction<T> importanceFunction, PropertyRunEvaluator<T> propertyEvaluator, int prolongation, bool collectDiagnostics)
			: base(run, collectDiagnostics)
		{
			System.Diagnostics.Debug.Assert(!(run is ZonePTASimulationRun<T>));
			if(importanceFunction == null) throw new ArgumentNullException(nameof(importanceFunction));
			if(propertyEvaluator == null) throw new ArgumentNullException(nameof(propertyEvaluator));
			if(prolongation < 0) throw new ArgumentOutOfRangeException(nameof(prolongation));

			this.importanceFunction = importanceFunction;
			this.propertyEvaluator = propertyEvaluator;
			this.prolongation = prolongation;
			this.lraPropertyEvaluator = propertyEvaluator as LongRunAverageValueProperty.LongRunAverageValueRunEvaluator<T>;
		}

		public override double[] PerformRun()
		{
			System.Diagnostics.Debug.Assert(count == 0);

			// Setup first main run
			result = 0.0;
			count = 1;
			Run.Start();
			if(Run.IsDone) return new double[] { propertyEvaluator.CurrentValue }; // no need to adjust results yet
			creationLevel[0] = 0; // main run never gets killed
			currentLevel[0] = initialStateLevel = importanceFunction.GetImportanceLevel(ref Run.CurrentState.State);
			splitLevel[0] = 0;
			System.Diagnostics.Debug.Assert(initialStateLevel >= 0);
			if(lraPropertyEvaluator != null)
			{
				// Advance the main run without splitting while in the transient phase
				while(lraPropertyEvaluator.IsTransient && !Run.IsDone) Run.Step();
				if(Run.IsDone) return new double[] { lraPropertyEvaluator.CurrentValue }; // no need to adjust results yet
				lraPropertyEvaluator.Reset();
				currentLevel[0] = importanceFunction.GetImportanceLevel(ref Run.CurrentState.State);
				if(currentLevel[0] > 0)
				{
					Split(1, currentLevel[0]);
					splitLevel[0] = currentLevel[0];
				}
			}
			else
			{
				// Split the main run according to the initial level
				Split(0, initialStateLevel);
			}

			// Perform main and child runs
			double mainRunDuration = lraPropertyEvaluator == null ? 1.0 : double.NaN;
			while(count != 0)
			{
				// Perform one run
				while(true) // simulation loop
				{
					Run.Step();
					if(Run.IsDone)
					{
						// Update result
						if(lraPropertyEvaluator == null)
						{
							double peResult = propertyEvaluator.CurrentValue;
							System.Diagnostics.Debug.Assert(!double.IsNaN(peResult));
							if(peResult != 0.0) result += GetWeight(splitLevel[0], peResult);
						}
						else
						{
							double currentReward = lraPropertyEvaluator.CurrentReward;
							if(currentReward != 0.0)
							{
								result += currentReward * GetWeight(splitLevel[0]);
								lraPropertyEvaluator.Reset();
							}
						}
						break; // the run has ended
					}
					else
					{
						// The run just made a jump
						int previousLevel = currentLevel[0];
						int newLevel = currentLevel[0] = importanceFunction.GetImportanceLevel(ref Run.CurrentState.State);
						System.Diagnostics.Debug.Assert(newLevel >= initialStateLevel);

						// Capture the reward in long-run average simulation
						if(lraPropertyEvaluator != null && previousLevel != newLevel)
						{
							double currentReward = lraPropertyEvaluator.CurrentReward;
							if(currentReward != 0.0)
							{
								result += currentReward * GetWeight(splitLevel[0]);
								lraPropertyEvaluator.Reset();
							}
						}

						// Split or kill
						if(newLevel > splitLevel[0])
						{
							// Split run
							Split(splitLevel[0] + 1, newLevel);
							splitLevel[0] = newLevel;
						}
						else if(newLevel < previousLevel)
						{
							if(creationLevel[0] == 0 && newLevel == 0)
							{
								// The main run returned to level 0: reset its last-split level
								splitLevel[0] = 0;
							}
							else if(newLevel < creationLevel[0] - prolongation || newLevel == 0)
							{
								// A child run went down far enough, or to level 0: kill it
								break;
							}
							else if(newLevel + prolongation < splitLevel[0])
							{
								// Adjust the last-split level
								splitLevel[0] = newLevel + prolongation;
							}
						}
					}
				}

				// If the main run just ended, get its duration for LRA evaluation
				if(double.IsNaN(mainRunDuration)) mainRunDuration = lraPropertyEvaluator.TotalEvaluationTime;

				// Update agenda
				if(--count != 0)
				{
					Run.CurrentState = agenda[count];
					agenda[count] = null; // cleanup
					currentLevel[0] = currentLevel[count];
					creationLevel[0] = creationLevel[count];
					splitLevel[0] = splitLevel[count];
				}
				System.Diagnostics.Debug.Assert(agenda[0] == null); // lives in Run.CurrentState instead
			}

			// Done
			return new double[] { result / mainRunDuration };
		}

		private void Split(int firstLevel, int lastLevel)
		{
			System.Diagnostics.Debug.Assert(count != 0);
			int runCount = 1;
			for(int intermediateLevel = firstLevel; intermediateLevel <= lastLevel; ++intermediateLevel)
			{
				int splittingFactor = importanceFunction.GetSplittingFactor(intermediateLevel);
				int newRunsOnThisLevel;
				try
				{
					checked
					{
						newRunsOnThisLevel = (splittingFactor - 1) * runCount;
						runCount += newRunsOnThisLevel;
					}
				}
				catch(OverflowException) { throw new ExplorationException("Too much splitting in a single jump (" + ((splittingFactor - 1) * (long)runCount).ToString(CI.InvariantCulture) + " new splits)."); }
				if(runCount > 262144) throw new ExplorationException("Too much splitting in a single run (" + (runCount - 1).ToString(CI.InvariantCulture) + " splits pending).");

				// Create new runs
				Reserve(newRunsOnThisLevel);
				for(int i = 0; i < newRunsOnThisLevel; ++i)
				{
					agenda[count] = Run.CurrentState.Clone();
					currentLevel[count] = lastLevel;
					creationLevel[count] = intermediateLevel;
					splitLevel[count] = lastLevel;
					++count;
				}
			}
		}

		private void Reserve(int newRunCount)
		{
			// Resize arrays, if necessary
			if(agenda.Length < count + newRunCount)
			{
				int newLength = Math.Max(agenda.Length * 2, count + newRunCount);
				Array.Resize(ref agenda, newLength);
				Array.Resize(ref currentLevel, newLength);
				Array.Resize(ref creationLevel, newLength);
				Array.Resize(ref splitLevel, newLength);
			}
		}

		[MethodImpl(MethodImplOptions.AggressiveInlining)]
		private double GetWeight(int level, double value = 1.0)
		{
			System.Diagnostics.Debug.Assert(level >= 0);
			while(level >= (lraPropertyEvaluator == null ? 0 : 1)) // divide incrementally to avoid errors since Math.Pow(splittingFactor, rareEventLevel) may be very large
			{
				value /= importanceFunction.GetSplittingFactor(level);
				--level;
			}
			return value;
		}

		public override double AverageRunLength
		{
			get { return double.NaN; }
		}

		public override string RunType { get { return Run.RunType + " / RESTART" + (prolongation == 0 ? string.Empty : "-P" + prolongation.ToString(CI.InvariantCulture) + " (old version)"); } }
	}
}
