﻿using System;
using System.Runtime.CompilerServices;
using Modest.Exploration;
using CI = System.Globalization.CultureInfo;

namespace Modest.Simulation
{
	sealed class RestartRunManager<T>: RunManager<T>
		where T : struct, IState<T>
	{
		private const int MaxPrecomputedWeights = 32;

		private readonly ImportanceFunction<T> importanceFunction;
		private readonly PropertyRunEvaluator<T> propertyEvaluator;
		private readonly int prolongation = 0;
		private readonly LongRunAverageValueProperty.LongRunAverageValueRunEvaluator<T> lraPropertyEvaluator;
		private readonly long[] levelFactors;

		private double result;

		public RestartRunManager(SimulationRun<T> run, ImportanceFunction<T> importanceFunction, PropertyRunEvaluator<T> propertyEvaluator, int prolongation, bool collectDiagnostics)
			: base(run, collectDiagnostics)
		{
			System.Diagnostics.Debug.Assert(!(run is ZonePTASimulationRun<T>));
			if(importanceFunction == null) throw new ArgumentNullException(nameof(importanceFunction));
			if(propertyEvaluator == null) throw new ArgumentNullException(nameof(propertyEvaluator));
			if(prolongation < 0) throw new ArgumentOutOfRangeException(nameof(prolongation));

			this.importanceFunction = importanceFunction;
			this.propertyEvaluator = propertyEvaluator;
			this.prolongation = prolongation;
			this.lraPropertyEvaluator = propertyEvaluator as LongRunAverageValueProperty.LongRunAverageValueRunEvaluator<T>;

			levelFactors = new long[Math.Min(MaxPrecomputedWeights, importanceFunction.TargetLevel + 1)];
			System.Diagnostics.Debug.Assert(importanceFunction.TargetLevel >= 0);
			levelFactors[0] = importanceFunction.Splits[0];
			for(int i = 1; i < Math.Min(MaxPrecomputedWeights, levelFactors.Length); ++i) levelFactors[i] = levelFactors[i - 1] * importanceFunction.GetSplittingFactor(i);
		}

		public override double[] PerformRun()
		{
			// Setup first main run
			result = 0.0;
			Run.Start();
			if(Run.IsDone) return new double[] { propertyEvaluator.CurrentValue }; // no need to adjust results yet
			if(lraPropertyEvaluator != null)
			{
				// Advance the main run without splitting while in the transient phase
				while(lraPropertyEvaluator.IsTransient && !Run.IsDone) Run.Step();
				if(Run.IsDone) return new double[] { lraPropertyEvaluator.CurrentValue }; // no need to adjust results yet
				lraPropertyEvaluator.Reset();
			}

			// Run
			Restart(0, importanceFunction.GetImportanceLevel(ref Run.CurrentState.State));

			// Return accumulated result, or the accumulated weighted reward divided by LRA evaluation phase duration
			return new double[] { result / (lraPropertyEvaluator == null ? 1.0 : lraPropertyEvaluator.TotalEvaluationTime) }; // the main run is the one executed last and the one that survives for the LRA evaluation duration, so we take its total time here
		}

		private void Restart(int restartLevel, int actualLevel)
		{
			int repetitions = importanceFunction.GetSplittingFactor(restartLevel);
			int savedActualLevel = actualLevel;
			var savedState = Run.CurrentState;
			while(repetitions-- > 0)
			{
				// Perform one run
				actualLevel = savedActualLevel;
				Run.CurrentState = repetitions == 0 ? savedState : savedState.Clone(); // the last repetition is this level's main run, so when it goes down, our caller will continue with its state
				while(true) // simulation loop
				{
					if(restartLevel < actualLevel)
					{
						// Split immediately
						Restart(restartLevel + 1, actualLevel);
						if(Run.IsDone) break; // the run has ended on a higher level (breaking here avoids one spurious call to Step on the next loop)
						actualLevel = importanceFunction.GetImportanceLevel(ref Run.CurrentState.State);
					}
					else
					{
						System.Diagnostics.Debug.Assert(restartLevel >= actualLevel);
						Run.Step();
						if(Run.IsDone)
						{
							// Update result
							if(lraPropertyEvaluator == null)
							{
								double peResult = propertyEvaluator.CurrentValue;
								System.Diagnostics.Debug.Assert(!double.IsNaN(peResult));
								if(peResult != 0.0) result += GetWeight(restartLevel, peResult);
							}
							else
							{
								double currentReward = lraPropertyEvaluator.CurrentReward;
								if(currentReward != 0.0) result += GetWeight(restartLevel, currentReward);
								lraPropertyEvaluator.Reset();
							}
							break; // the run has ended
						}
						else
						{
							// The run just made a jump
							actualLevel = importanceFunction.GetImportanceLevel(ref Run.CurrentState.State);

							// Capture the reward in long-run average simulation
							if(lraPropertyEvaluator != null && restartLevel != actualLevel)
							{
								double currentReward = lraPropertyEvaluator.CurrentReward;
								if(currentReward != 0.0) result += GetWeight(restartLevel, currentReward);
								lraPropertyEvaluator.Reset();
							}

							// Split or kill
							if(restartLevel < actualLevel)
							{
								// Split run
								Restart(restartLevel + 1, actualLevel);
								if(Run.IsDone) break; // the run has ended on a higher level (breaking here avoids one spurious call to Step on the next loop)
								actualLevel = importanceFunction.GetImportanceLevel(ref Run.CurrentState.State);
							}
							else if(actualLevel == 0 && restartLevel != 0 || actualLevel < restartLevel - prolongation)
							{
								// A child run went down far enough, or to level 0: kill it
								break;
							}
						}
					}
				}
			}
		}

		[MethodImpl(MethodImplOptions.AggressiveInlining)]
		private double GetWeight(int level, double value)
		{
			System.Diagnostics.Debug.Assert(level >= 0);
			while(level >= MaxPrecomputedWeights) // divide incrementally to avoid errors since Math.Pow(splittingFactor, level) may be very large
			{
				value /= importanceFunction.GetSplittingFactor(level);
				--level;
			}
			return value / levelFactors[level];
		}

		public override double AverageRunLength
		{
			get { return double.NaN; }
		}

		public override string RunType { get { return Run.RunType + " / RESTART" + (prolongation == 0 ? string.Empty : "-P" + prolongation.ToString(CI.InvariantCulture)); } }
	}
}
