#!/usr/bin/env bash

[ -f ./utils.sh ] || (echo "File utils.sh not found; aborting" && exit 1);
source ./utils.sh;
assert_fig;

MODELS_DIR="./models";
MODELS=(
	"2tandem_queue_Jackson.sa"
	"3tandem_queue_Jackson.sa"
	"3tandem_queue_nonJackson.sa"
	"7nodes_network_Jackson.sa"
	"7nodes_network_nonJackson.sa"
);
for M in "${MODELS[@]}"; do
	assert_file "${MODELS_DIR}/$M";
done
# Define one (ad hoc) Importance Function per model above; in order!
IFUNS=(
	'63*q1+100*q2'
	'37*q1+63*q2+100*q3'
	'13*q1+50*q2+100*q3'
	'23*(n1+n2+n3+n4)+45*(n5+n6)+100*n7;0;169200'
	'23*(n1+n2+n3+n4)+45*(n5+n6)+100*n7;0;169200'
);
# Define one (ad hoc) set of thresholds per model above; in order!
THRESHOLDS=(
	"200:3,300:2,400:2,500:3,600:3,700:3,800:2,900:3,1000:3,1100:3,1200:3,1300:3,1400:2,1500:3,1600:3,1700:3,1800:3,1900:3,2000:3,2100:3,2200:3,2300:3,2400:3,2500:3,2600:3,2700:3,2800:3,2900:5"
	"200:2,300:2,400:2,500:2,600:2,700:3,800:2,900:3,1000:3,1100:3,1200:2,1300:3,1400:3,1500:3,1600:2,1700:3,1800:3,1900:3,2000:2,2100:3,2200:3,2300:3,2400:3,2500:3,2600:2,2700:3,2800:3,2900:3,3000:3,3100:2,3200:3,3300:3,3400:3"
    "150:2,250:3,300:2,350:3,400:2,450:3,500:3,550:3,600:3,650:3,700:3,750:3,800:3,850:3,900:3,950:3,1000:3,1050:3,1100:3,1150:3,1200:3,1250:3,1300:3,1350:3,1400:4,1450:3,1500:3,1550:4,1600:5"
	"200:2,300:2,400:3,500:3,600:3,700:3,800:3,900:3,1000:3,1100:3,1200:3,1300:3,1400:3,1500:3,1600:3,1700:3,1800:3,1900:4,2000:3,2100:3,2200:3,2300:3,2400:3,2500:3,2600:3,2700:3,2800:3,2900:3,3000:3,3100:3,3200:2"
	"200:2,300:2,400:2,500:3,600:3,700:3,800:3,900:3,1000:3,1100:3,1200:2,1300:3,1400:3,1500:3,1600:3,1700:2,1800:3,1900:4,2000:3,2100:3,2200:2,2300:3,2400:3,2500:3,2600:3,2700:2,2800:3,2900:4"
);

if [ -z ${TEST_RUN+x} ] || [ $TEST_RUN -ne 1 ]; then
	# Full experimental reproduction
	NUM_REPETITIONS=5;
	# Experimentation modes
	MODES=(
		"light"
		"heavy"
	);
	# Models to test with demanding criteria (more engines, tighter CI)
	MODELS_IDX_HEAVY=(
		0
	);
	# Models to test in general experiments (needn't be disjoint with above)
	MODELS_IDX_LIGHT=(
		1
		2
		3
		4
	);
	declare -A MODELS_IDX_ALL=(
		[${MODES[0]}]="${MODELS_IDX_LIGHT[@]}"
		[${MODES[1]}]="${MODELS_IDX_HEAVY[@]}"
	)
	# RESTART variants to compare against standard RESTART in general experiments
	ENGINES_LIGHT=(
		"restart0"
		"restart1"
		"restart2"
	);
	# RESTART variants to try with (first) smaller experiments
	ENGINES_HEAVY=(
		"restart0"
		"restart1"
		"restart2"
		"restart3"
		"restart4"
	);
	declare -A ENGINES_ALL=(
		[${MODES[0]}]="${ENGINES_LIGHT[@]}"
		[${MODES[1]}]="${ENGINES_HEAVY[@]}"
	)
	# Confidence Interval stop criterion (conf.lvl. and precision = 2*rel.err.)
	CI_LIGHT="--stop-conf .95 .2";
	CI_HEAVY="--stop-conf .95 .01";
	declare -A CI_ALL=(
		[${MODES[0]}]="${CI_LIGHT[@]}"
		[${MODES[1]}]="${CI_HEAVY[@]}"
	)
else
	# Fast test run
	NUM_REPETITIONS=1;
	MODES=("test");
	declare -A MODELS_IDX_ALL=([${MODES[0]}]="0 1 2 3 4");
	declare -A ENGINES_ALL=([${MODES[0]}]="restart0 restart2");
	declare -A CI_ALL=([${MODES[0]}]="--stop-conf .95 .4");
fi

# Extra command line options to fig
extra() { echo "--parlare 1 --timeout 15h --rng-seed $((REP+1))"; }

export RESULTS_DIR="./results/FIG";
mkdir -p ${RESULTS_DIR};
RESULTS=`make_fresh_results_dir`;
echo "Saving results in $RESULTS";

# Experimentation
for EM in "${MODES[@]}"; do
	echo "Running experiments in mode \"${EM}\"";
	for (( REP=0 ; REP<NUM_REPETITIONS ; REP++ )); do
		echo "> Repetition $((REP+1)) out of ${NUM_REPETITIONS}";
		CI="${CI_ALL[$EM]}";
		ENGINES=(${ENGINES_ALL[$EM]});
		MODELS_IDX=(${MODELS_IDX_ALL[$EM]});
		for i in "${MODELS_IDX[@]}"; do
			MNAME="${MODELS[i]%%.sa}";
			echo "  - Experiments for ${MNAME}";
			MODEL="${MODELS_DIR}/${MODELS[i]}";
			IFUN="--adhoc '${IFUNS[i]}'";  # enclose in single quotes!
			THRE="--thresholds-adhoc ${THRESHOLDS[i]}";
			for ENG in "${ENGINES[@]}"; do
				LOG="${RESULTS}/${REP}_${MNAME}_${ENG}";
				poll_till_free_FIG "\-\-adhoc";
				echo "    - $ENG";
				fig $MODEL $CI $IFUN $THRE -e $ENG `extra` 1>"$LOG.out" \
				                                           2>"$LOG.err" &
			done # this engine (aka RESTART-Pj variant)
		done # this model
	done # this repetition
done # this experiment mode

wait;

echo "Finished experiments: logs stored in $RESULTS/*";
echo "Compiling summary tables";

# Results summary
CHART="$RESULTS/summary.txt";
cat "summary_header.txt" > "$CHART";
for EM in "${MODES[@]}"; do
	ENGINES=(${ENGINES_ALL[$EM]});
	MODELS_IDX=(${MODELS_IDX_ALL[$EM]});
	for i in "${MODELS_IDX[@]}"; do
		MNAME="${MODELS[i]%%.sa}";
		CI="`echo ${CI_ALL[$EM]} | grep -o "[0-9\.\ ]\+"`";
		echo "$MNAME  |  CI: $CI" >> $CHART;
		for ENG in "${ENGINES[@]}"; do
			echo >> "$CHART";
			for (( REP=0 ; REP<NUM_REPETITIONS ; REP++ )); do
				LOG="${RESULTS}/${REP}_${MNAME}_${ENG}.out";
				ESTIMAT="`awk '/Computed estimate/{ print $4    }' $LOG`";
				SAMPLES="`awk '/Computed estimate/{ print $5")" }' $LOG`";
				RUNTIME="`awk '/Estimation time/  { print $4    }' $LOG`";
				echo -e "$ENG\t $ESTIMAT\t $SAMPLES\t\t $RUNTIME" >> $CHART;
			done # this repetition
		done # this engine (aka RESTART variant)
		echo >> "$CHART";
	done # this model
done # this experiment mode
mkdir -p "$RESULTS/logs";
mv "$RESULTS"/*.{out,err} "$RESULTS/logs/";

echo "Summary tables saved in $PWD/$CHART";
echo "Finished experimental reproduction on `date "+%b %d @ %T"`";

exit 0;
