#!/usr/bin/env bash

[ -f ./utils.sh ] || (echo "File utils.sh not found; aborting" && exit 1);
source ./utils.sh;
assert_file RESTART_Jose.cpp  # C-code source file by J. Villen-Altamirano
assert_file Makefile

if [ -z ${TEST_RUN+x} ] || [ $TEST_RUN -ne 1 ]; then
	# Full experimental reproduction
	NUM_REPETITIONS=5;
	MANY_P=4;
	FEW_P=2;
	STEP=1;
else
	# Fast test run
	NUM_REPETITIONS=1;
	MANY_P=2;
	FEW_P=2;
	STEP=2;
fi

# Make these match the rules of Makefile, e.g. 2TANQ_J_RESTART
tandem2_J=( `seq -f "2TANQ_J_RESTART%g"  0 $STEP $MANY_P`);  # RESTART-P0 to -P4
tandem3_J=( `seq -f "3TANQ_J_RESTART%g"  0 $STEP $FEW_P`);
tandem3_nJ=(`seq -f "3TANQ_nJ_RESTART%g" 0 $STEP $FEW_P`);
nodes7_J=(  `seq -f "7NODN_J_RESTART%g"  0 $STEP $FEW_P`);
nodes7_nJ=( `seq -f "7NODN_nJ_RESTART%g" 0 $STEP $FEW_P`);
# Array that contains the arrays defined above
MODELS=(
	tandem2_J
	tandem3_J
	tandem3_nJ
	nodes7_J
	nodes7_nJ
);
# More descriptive names for the models
MODELS_NAMES=(
	"2tandem_queue_Jackson"
	"3tandem_queue_Jackson"
	"3tandem_queue_nonJackson"
	"7nodes_network_Jackson"
	"7nodes_network_nonJackson"
)
# Confidence level and relative error for each model (hard-coded in source file)
CI=(
	"95% .005"
	"95% .1"
	"95% .1"
	"95% .1"
	"95% .1"
)  # NOTE: CI precision == 2*Rel.Err

export RESULTS_DIR="./results/Jose";
mkdir -p ${RESULTS_DIR};
RESULTS=`make_fresh_results_dir`;
echo "Saving results in $RESULTS";

make clean 2>&1 >/dev/null;

# For as many repetitions as desired
for (( REP=0 ; REP<NUM_REPETITIONS ; REP++ )); do
	echo "> Repetition $((REP+1)) out of ${NUM_REPETITIONS}";
	# For each model
	for (( i=0 ; i<${#MODELS[@]} ; i++ )); do
		echo "  - Experiments for ${MODELS_NAMES[i]}";
		VARIANTS=("${MODELS[i]}[@]");
		# For each RESTART variant
		for MV in ${!VARIANTS}; do
			LOG="${RESULTS}/${REP}_${MV}";
			make $MV RNGSEED=$REP 2>&1 >/dev/null;
			poll_till_free_RESTART;
			echo "    - ${MV##*_}";  # RESTART variant, e.g. RESTART1
			./$MV 1> ${LOG}.out 2> ${LOG}.err &
		done # this RESTART variant
	done # this model
	sleep 1;
	make clean 2>&1 >/dev/null;  # recompile to change RNG seed in next REP
done # this repetition

wait;

echo "Finished experiments: logs stored in $RESULTS/*";
echo "Compiling summary tables";

# Results summary
CHART="$RESULTS/summary.txt";
cat "summary_header.txt" > "$CHART";
for (( i=0 ; i<${#MODELS[@]} ; i++ )); do
	echo "${MODELS_NAMES[i]}  |  CI: ${CI[i]}" >> $CHART;
	VARIANTS=("${MODELS[i]}[@]");
	for MV in ${!VARIANTS}; do
		echo >> "$CHART";
		VARIANT="${MV##*_}";  # RESTART variant, e.g. RESTART1
		for (( REP=0 ; REP<NUM_REPETITIONS ; REP++ )); do
			LOG="${RESULTS}/${REP}_${MV}.out";
			ESTIMAT="`awk '/Estimador puntual/     { print $4       }' $LOG`";
			SAMPLES="`awk '/Numero de simulaciones/{ print "("$5")" }' $LOG`";
			RUNTIME="`awk '/Tiempo ejecucion/      { print $5       }' $LOG`";
			echo -e "$VARIANT\t $ESTIMAT\t $SAMPLES\t\t $RUNTIME" >> $CHART;
		done # this repetition
	done # this RESTART variant
	echo >> "$CHART";
done # this model
mkdir -p "$RESULTS/logs";
mv "$RESULTS"/*.{out,err} "$RESULTS/logs/";

echo "Summary tables saved in $PWD/$CHART";
echo "Finished experimental reproduction on `date "+%b %d @ %T"`";

exit 0;
